"""
Pypeline's built-in markups; use these as examples for how to create your
own markup functions.
"""



try:
    import creoleparser
except ImportError:
    pass
else:
    def creole_markup():
        return r'creole', lambda s: creoleparser.text2html(s)


try:
    import markdown
except ImportError:
    pass
else:
    def markdown_markup(extensions=None):
        return r'md|mkdn?|mdown|markdown', lambda s: markdown.markdown(s, extensions=extensions or [])

try:
    import textile
except ImportError:
    pass
else:
    def textile_markup():
        return r'textile', lambda s: textile.textile_restricted(s, lite=False, noimage=False)

try:
    from docutils.core import publish_parts
    from docutils.writers.html4css1 import Writer
except ImportError:
    pass
else:
    def rest_markup():
        # see http://docutils.sourceforge.net/docs/howto/security.html
        settings = dict(
            cloak_email_addresses=True,
            file_insertion_enabled=False,
            raw_enabled=False,
            strip_comments=True,
            doctitle_xform=False,
            report_level=5,
        )

        def render(s):
            parts = publish_parts(s, writer=Writer(), settings_overrides=settings)
            if 'html_body' in parts:
                html = parts['html_body']
                return html
            return ''
        return r're?st(\.txt)?', render



from allpairspy import AllPairs


class TestPairwise:
    def __init__(self):
        self.data = {}
    
    def test(self):
        # run all '_test*' methods
        tests = [f for f in dir(self) if f.startswith('_test')]
        if not tests:
            return
        pairwise = AllPairs([tests, list(self.data.keys())], filter_func=self.is_valid_combination)

        # code to make allpairs and nose play nicely with each 
        # other; maybe wrap in a nose plugin some day?
        for (test_func, data) in pairwise:
            # http://somethingaboutorange.com/mrl/projects/nose/0.11.1/writing_tests.html#test-generators
            def named_test(*args, **kwargs):
                if not getattr(self,'app',None):
                    self.setUp()
                return getattr(self,test_func)(*args, **kwargs)
            named_test.description = (self.__class__.__module__
                                      + '.' + self.__class__.__name__
                                      + '.' + test_func
                                      + '(%s)' % ','.join([data,]))
            #named_test.setup = self.setUp
            #named_test.teardown = self.tearDown
            yield named_test, data
    
    def is_valid_combination(self, row):
        return True




import cgi
import re
from types import FunctionType

from functools import partial

import six
import bleach
from bleach.linkifier import LinkifyFilter
from bleach.css_sanitizer import CSSSanitizer
import html5lib.filters.sanitizer

# based on allura.lib.utils.ForgeHTMLSanitizerFilter
_form_tags = {'button', 'datalist', 'fieldset', 'form', 'input', 'label', 'legend', 'meter', 'optgroup', 'option',
              'output', 'progress', 'select', 'textarea', }
_ns_html = html5lib.constants.namespaces['html']
_extra_allowed_elements = set([
            (_ns_html, 'summary'),
        ])
ALLOWED_TAGS = [
    el[1]  # just the tag name, no namespace
    # for el in html5lib.filters.sanitizer.allowed_elements
    for el in set(html5lib.filters.sanitizer.allowed_elements) | _extra_allowed_elements
    if el[0] == _ns_html
    and el[1] not in _form_tags
]

# copied from html5lib.filters.sanitizer.allowed_attributes but only the HTML ones at tope
# TODO: delineate by tag, instead of '*' for all
# bleach_whitelist.markdown_attrs and .print_attrs has a very short list
# bleach.ALLOWED_ATTRIBUTES is just an example, not realistic
ALLOWED_ATTRIBUTES = {
    '*': [
        el[1] for el in [
            (None, 'abbr'),
            (None, 'accept'),
            (None, 'accept-charset'),
            (None, 'accesskey'),
            (None, 'action'),
            (None, 'align'),
            (None, 'alt'),
            (None, 'autocomplete'),
            (None, 'autofocus'),
            (None, 'axis'),
            (None, 'background'),
            (None, 'balance'),
            (None, 'bgcolor'),
            (None, 'bgproperties'),
            (None, 'border'),
            (None, 'bordercolor'),
            (None, 'bordercolordark'),
            (None, 'bordercolorlight'),
            (None, 'bottompadding'),
            (None, 'cellpadding'),
            (None, 'cellspacing'),
            (None, 'ch'),
            (None, 'challenge'),
            (None, 'char'),
            (None, 'charoff'),
            (None, 'choff'),
            (None, 'charset'),
            (None, 'checked'),
            (None, 'cite'),
            (None, 'class'),
            (None, 'clear'),
            (None, 'color'),
            (None, 'cols'),
            (None, 'colspan'),
            (None, 'compact'),
            (None, 'contenteditable'),
            (None, 'controls'),
            (None, 'coords'),
            (None, 'data'),
            (None, 'datafld'),
            (None, 'datapagesize'),
            (None, 'datasrc'),
            (None, 'datetime'),
            (None, 'default'),
            (None, 'delay'),
            (None, 'dir'),
            (None, 'disabled'),
            (None, 'draggable'),
            (None, 'dynsrc'),
            (None, 'enctype'),
            (None, 'end'),
            (None, 'face'),
            (None, 'for'),
            (None, 'form'),
            (None, 'frame'),
            (None, 'galleryimg'),
            (None, 'gutter'),
            (None, 'headers'),
            (None, 'height'),
            (None, 'hidefocus'),
            (None, 'hidden'),
            (None, 'high'),
            (None, 'href'),
            (None, 'hreflang'),
            (None, 'hspace'),
            (None, 'icon'),
            (None, 'id'),
            (None, 'inputmode'),
            (None, 'ismap'),
            (None, 'keytype'),
            (None, 'label'),
            (None, 'leftspacing'),
            (None, 'lang'),
            (None, 'list'),
            (None, 'longdesc'),
            (None, 'loop'),
            (None, 'loopcount'),
            (None, 'loopend'),
            (None, 'loopstart'),
            (None, 'low'),
            (None, 'lowsrc'),
            (None, 'max'),
            (None, 'maxlength'),
            (None, 'media'),
            (None, 'method'),
            (None, 'min'),
            (None, 'multiple'),
            (None, 'name'),
            (None, 'nohref'),
            (None, 'noshade'),
            (None, 'nowrap'),
            (None, 'open'),
            (None, 'optimum'),
            (None, 'pattern'),
            (None, 'ping'),
            (None, 'point-size'),
            (None, 'poster'),
            (None, 'pqg'),
            (None, 'preload'),
            (None, 'prompt'),
            (None, 'radiogroup'),
            (None, 'readonly'),
            (None, 'rel'),
            (None, 'repeat-max'),
            (None, 'repeat-min'),
            (None, 'replace'),
            (None, 'required'),
            (None, 'rev'),
            (None, 'rightspacing'),
            (None, 'rows'),
            (None, 'rowspan'),
            (None, 'rules'),
            (None, 'scope'),
            (None, 'selected'),
            (None, 'shape'),
            (None, 'size'),
            (None, 'span'),
            (None, 'src'),
            (None, 'start'),
            (None, 'step'),
            (None, 'style'),
            (None, 'summary'),
            (None, 'suppress'),
            (None, 'tabindex'),
            (None, 'target'),
            (None, 'template'),
            (None, 'title'),
            (None, 'toppadding'),
            (None, 'type'),
            (None, 'unselectable'),
            (None, 'usemap'),
            (None, 'urn'),
            (None, 'valign'),
            (None, 'value'),
            (None, 'variable'),
            (None, 'volume'),
            (None, 'vspace'),
            (None, 'vrml'),
            (None, 'width'),
            (None, 'wrap'),
        ]
    ],
}

# less than html5lib.filters.sanitizer.allowed_protocols
# more than bleach.sanitizer.ALLOWED_PROTOCOLS
ALLOWED_PROTOCOLS = ['http', 'https', 'mailto', 'ftp', 'irc', 'ircs', 'nntp', 'sftp']

RE_MATCH_NOTHING = re.compile("a^")

CSS_SANITIZER = CSSSanitizer()


class Markup:

    def __init__(self, *markup_pkgs_or_fns, **markup_extensions):
        # Define markups
        self.markups = []
        for markup_pkg in markup_pkgs_or_fns:
            if type(markup_pkg) == FunctionType:
                self.add_renderer(markup_pkg)
            else:
                renderers = [r for r in dir(markup_pkg) if r.endswith('_markup')]
                for renderer in renderers:
                    func = getattr(markup_pkg, renderer)
                    self.add_renderer(func, markup_extensions)


        self.bleach_cleaner = bleach.Cleaner(
            tags=ALLOWED_TAGS,
            attributes=ALLOWED_ATTRIBUTES,
            css_sanitizer=CSS_SANITIZER,
            protocols=ALLOWED_PROTOCOLS,
            # LinkifyFilter is used to turn otherwise plaintext into <a> tags. We don't want that, but we do
            #   want its side-effect of 'callbacks' which can alter the <a> tags both already in the file and newly
            #   generated by linkify. We want the default side-effects which add 'rel="nofollow"' to all <a>.
            #   We set url_re to a regex that should never match any text ever to prevent linkify from generating new
            #   <a> from plaintext -- so the callbacks only affects the already existing <a> tags.
            filters=[partial(LinkifyFilter, callbacks=bleach.linkifier.DEFAULT_CALLBACKS, url_re=RE_MATCH_NOTHING)],
            strip=False,  # escape unsafe tags instead
            strip_comments=True,
        )

    @property
    def markups_names(self):
        return [m['name'] for m in self.markups]

    def add_renderer(self, func, markup_extensions):
        r'''
        Add a function that is responsible for rendering a particular
        markup syntax to HTML.

        The function name should end in \*_markup. The
        function must return a tuple of the pattern used to match supported
        file extensions and a callback to render a string of text.

        The callback recieves one parameter, a Unicode string. It should
        return an HTML representation of that string. Simple callbacks can be
        lambdas, such as creole, textile, or markup.
        '''
        name = func.__name__[:-len('_markup')]

        try:
            ext_pattern, render_fn = func(markup_extensions.get(name))
        except TypeError:
            ext_pattern, render_fn = func()

        self.markups.append(dict(name = name,
                                 ext_pattern = ext_pattern,
                                 render_fn = render_fn,
                            ))

    def render(self, filename, content=None):
        '''
        Provided a filename or a filename and content, render the
        content from a particular markup syntax to HTML. The markup
        syntax is chosen based on matching patterns (e.g., extensions)
        with the filename.
        '''
        if not content:
            with open(filename, 'rb') as f:
                content = f.read()
        content = self.unicode(content)
        (proc, name) = self.renderer(filename)
        if proc:
            html = proc(content)
            return self.bleach_cleaner.clean(self.unicode(html))
        else:
            # make sure unrendered content is at least escaped
            return cgi.escape(content)

    def unicode(self, content):
        '''
        Normalize a variety of encodings into unicode.
        '''
        if type(content) == str:
            return content
        try:
            content.decode('ascii')
        except UnicodeDecodeError: #pragma: nocover
            # try 2 common encodings
            try:
                content = content.decode('utf-8')
            except UnicodeDecodeError:
                try:
                    content = content.decode('latin_1')
                except UnicodeDecodeError:
                    raise Exception('The content uses an unsupported encoding.')
        return str(content)

    def can_render(self, filename):
        '''
        Check to see if a particular file is supported. If the file is
        supported, the name of the markup is returned. Otherwise, the
        function returns None.
        '''
        (proc, name) = self.renderer(filename)
        return name

    def renderer(self, filename):
        '''
        Search for any markups that are responsible for the provided
        filename. Returns the function used to render the file and
        the name of the markup.
        '''
        for markup in self.markups:
            if re.search(r"\.(%s)$" % markup['ext_pattern'], filename):
                return markup['render_fn'], markup['name']
        return None, None

from pypeline import markups
markup = Markup(markups)


from setuptools import setup, find_packages

version = '0.6.0'

setup(name='Pypeline',
      version=version,
      description="Easy rendering of markup languages",
      long_description="""Provides an easy, pluggable way to support rendering an arbitrary markup syntax (ReST, Markdown, etc.) to HTML.
""",
      classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: CGI Tools/Libraries',
        'Topic :: Software Development :: Documentation',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Filters',
        'Topic :: Text Processing :: Markup :: HTML',
        ], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='markup, markdown, textile, creole, text',
      author='Kyle Adams',
      author_email='kyle@geek.net',
      url='http://pypeline.sourceforge.net',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=True,
      python_requires='>=3.7',
      install_requires=[
        'bleach[css]>=5',
        'html5lib',
      ],
      extras_require={
        'creole': ["Creoleparser >= 0.7.2"],
        'markdown': ["Markdown >= 2.0.3"],
        'textile': ["textile >= 2.1.4"],
        'rst': ["docutils >= 0.7"],
      },
      tests_require=[
        'allpairspy',
        'nose',
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      test_suite='nose.collector',
      )



import os
import os.path

from bleach.css_sanitizer import ALLOWED_CSS_PROPERTIES

from nose.tools import assert_equal, assert_not_equal
from nose import SkipTest

from pypeline.markup import markup
from pypeline.markup import ALLOWED_TAGS, ALLOWED_ATTRIBUTES
from pypeline.tests import TestPairwise
import six

from unittest import TestCase
TestCase.maxDiff = None

class TestMarkup(TestPairwise):

    basedir = os.path.join(os.path.dirname(__file__), 'markups')

    """
    test_* methods are run like regular tests
    _test_* methods are run with the allpairs logic (e.g. once for each readme)
        and that is kicked off by test()
    """
    def __init__(self):
        super().__init__()
        self.data = {}
        files = os.listdir(self.basedir)
        for f in files:
            format = os.path.splitext(f)[1].lstrip('.')
            if format == 'html':
                continue
            self.data[format] = f

    def setUp(self):
        pass

    def _test_rendering(self, format):
        if format not in markup.markups_names and format != 'plaintext':
            raise SkipTest()
        readme = self.data[format]
        source_file = open(os.path.join(self.basedir, readme), encoding='utf-8')
        source = source_file.read()
        expected_file = open(os.path.join(self.basedir, '%s.html' % readme), encoding='utf-8')
        expected = expected_file.read()
        actual = markup.render(os.path.join(self.basedir, readme))
        #assert_true(isinstance(actual, unicode))
        if source != expected:
            assert_not_equal(source, actual, "Did not render anything.")
        assert_equal(expected, actual)

    def test_can_render(self):
        assert_equal(None, markup.can_render('README.cmd'))
        if 'markdown' not in markup.markups_names:
            raise SkipTest()
        else:
            assert_equal('markdown', markup.can_render('README.markdown'))
            assert_equal('markdown', markup.can_render('README.md'))

    def test_unicode_utf8(self):
        chinese = markup.unicode('華語')
        assert_equal(chinese, '華語')
        assert_equal(type(chinese), str)

    def test_unicode_ascii(self):
        ascii = markup.unicode('abc')
        assert_equal(ascii, 'abc')
        assert_equal(type(ascii), str)

    def test_unicode_latin1(self):
        latin1 = 'abcdé'.encode('latin_1')
        latin1 = markup.unicode(latin1)
        assert_equal(latin1, 'abcdé')
        assert_equal(type(latin1), str)


class TestHTMLAllowedValues:

    def test_ALLOWED_TAGS(self):
        assert 'pre' in ALLOWED_TAGS
        assert 'button' not in ALLOWED_TAGS
        assert 'glyph' not in ALLOWED_TAGS  # svg

    def test_ALLOWED_ATTRIBUTES(self):
        assert 'alt' in ALLOWED_ATTRIBUTES['*']
        assert 'clip-path' not in ALLOWED_ATTRIBUTES['*']  # svg

    def test_ALLOWED_CSS_PROPERTIES(self):
        assert 'color' in ALLOWED_CSS_PROPERTIES
        assert 'position' not in ALLOWED_CSS_PROPERTIES
        assert '-moz-binding' not in ALLOWED_CSS_PROPERTIES